/**
  *  \file server/jsonin.cc
  *  \brief Server: JSON input
  */

#include <memory>
#include "server/jsonin.h"
#include "int/array.h"
#include "int/interror.h"

static bool
isIdentifier(char c)
{
    return (c >= 'a' && c <= 'z')
        || (c >= 'A' && c <= 'Z')
        || (c >= '0' && c <= '9')
        || c == '_';
}



JsonTokenizer::JsonTokenizer(string_t text)
    : current_token(kEOF),
      current_float(0),
      current_int(0),
      current_string(),
      text(text),
      pos(0)
{ }

JsonTokenizer::Token
JsonTokenizer::readNext()
{
    /* Skip whitespace */
    while (pos < text.size() && (text[pos] == ' ' || text[pos] == '\r' || text[pos] == '\n' || text[pos] == '\t'))
        ++pos;

    /* End reached? */
    if (pos >= text.size()) {
        current_token = kEOF;
        return current_token;
    }

    /* What is it? */
    switch (text[pos]) {
     case '[':
        current_token = kStartArray;
        ++pos;
        return current_token;
     case ']':
        current_token = kEndArray;
        ++pos;
        return current_token;
     case '{':
        current_token = kStartHash;
        ++pos;
        return current_token;
     case '}':
        current_token = kEndHash;
        ++pos;
        return current_token;
     case ':':
        current_token = kColon;
        ++pos;
        return current_token;
     case ',':
        current_token = kComma;
        ++pos;
        return current_token;
     case '"':
        current_token = kString;
        current_string.clear();
        ++pos;
        while (pos < text.size() && text[pos] != '"') {
            if (text[pos] == '\\') {
                ++pos;
                if (pos >= text.size()) {
                    current_token = kError;
                    return current_token;
                }
                switch (text[pos]) {
                 case '\\':
                 case '"':
                 case '/':
                    current_string += text[pos];
                    break;
                 case 't':
                    current_string += '\t';
                    break;
                 case 'n':
                    current_string += '\n';
                    break;
                 default:
                    /* FIXME: JSON specs \b \f \t \uXXXX, but we don't currently use those. */
                    pos = text.size();
                    current_token = kError;
                    return current_token;
                }
                ++pos;
            } else {
                current_string += text[pos];
                ++pos;
            }
        }
        if (pos >= text.size()) {
            current_token = kError;
        } else {
            ++pos;
        }
        return current_token;
    }

    /* Simple cases have been processes, what remains are numbers and keywords. */
    if (text[pos] == '-' || (text[pos] >= '0' && text[pos] <= '9')) {
        /*
         *  Numbers are most complicated. Differences to spec:
         *  - we don't handle exponential notation
         *  - we don't refuse "octal" (i.e. "0377", or even "000000123.4")
         */

        /* Read sign */
        bool negate = false;
        if (text[pos] == '-') {
            negate = true;
            ++pos;
        }

        /* Read integer */
        bool isInt = true;
        bool hadDigits = false;
        current_int = 0;
        current_float = 0;
        while (pos < text.size() && text[pos] >= '0' && text[pos] <= '9') {
            hadDigits = true;
            current_float = 10*current_float + (text[pos] - '0');
            if (isInt && current_float < 2147483648.0) {
                current_int = 10*current_int + (text[pos] - '0');
            } else {
                isInt = false;
            }
            ++pos;
        }

        /* Read fraction */
        if (hadDigits && pos < text.size() && text[pos] == '.') {
            isInt = false;
            double den = 1.0;
            double num = 0.0;
            ++pos;
            while (pos < text.size() && text[pos] >= '0' && text[pos] <= '9') {
                hadDigits = true;
                num = 10*num + (text[pos] - '0');
                den *= 10;
                ++pos;
            }
            current_float += num/den;
        }

        /* Return result */
        if (!hadDigits) {
            pos = text.size();
            current_token = kError;
        } else if (isInt) {
            current_token = kInt;
            if (negate)
                current_int = -current_int;
        } else {
            current_token = kFloat;
            if (negate)
                current_float = -current_float;
        }
        return current_token;
    }

    if (isIdentifier(text[pos])) {
        size_t length = 1;
        while (pos + length < text.size() && isIdentifier(text[pos + length]))
            ++length;
        current_string.assign(text, pos, length);
        current_token = kKeyword;
        pos += length;
        return current_token;
    }

    /* Nothing recognized */
    current_token = kError;
    pos = text.size();
    return current_token;
}



/** Read a JSON array tree into a data structure.
    Assumes tok looking at first token.
    Returns the parsed data.
    Throws IntError on parse problems. */
IntValue*
readTree(JsonTokenizer& tok)
{
    std::auto_ptr<IntValue> result;
    switch (tok.getCurrent()) {
     case JsonTokenizer::kString:
        result.reset(makeStringValue(tok.getStringValue()));
        tok.readNext();
        break;
        
     case JsonTokenizer::kKeyword:
        if (tok.getStringValue() == "null") {
            /* leave result null */
        } else if (tok.getStringValue() == "false") {
            result.reset(makeBoolValue(false));
        } else if (tok.getStringValue() == "true") {
            result.reset(makeBoolValue(true));
        } else {
            throw IntError("Unknown keyword");
        }
        tok.readNext();
        break;
        
     case JsonTokenizer::kInt:
        result.reset(makeIntValue(tok.getIntValue()));
        tok.readNext();
        break;

     case JsonTokenizer::kFloat:
        result.reset(makeFloatValue(tok.getFloatValue()));
        tok.readNext();
        break;

     case JsonTokenizer::kStartArray:
        {
            Ptr<IntArrayData> data = new IntArrayData();
            tok.readNext();
            while (tok.getCurrent() != JsonTokenizer::kEndArray) {
                data->content.pushNew(readTree(tok));
                if (tok.getCurrent() == JsonTokenizer::kEndArray)
                    break;
                if (tok.getCurrent() != JsonTokenizer::kComma)
                    throw IntError("Expecting ','");
                tok.readNext();
            }
            tok.readNext();
            data->addDimension(data->content.size());
            result.reset(new IntArray(data));
        }
        break;

     default:
        throw IntError("Parse error");
    }

    return result.release();
}
